\ Extensions to SOD Forth kernel to make a complete Forth system.
\ Copyright 1994 L.C. Benschop Eindhoven, The Netherlands.
\ The program is released under the GNU General Public License version 2.
\ There is NO WARRANTY.

\ Dec. 5 '94: mhx added TIMER-RESET .ELAPSED PAGE CLS


: \G	POSTPONE \ ; IMMEDIATE
\G comment til end of line for inclusion in glossary.

\ PART 1: MISCELLANEOUS WORDS.

: COMPARE ( addr1 u1 addr2 u2 -- diff ) 
\G Compare two strings. diff is negative if addr1 u1 is smaller, 0 if it 
\G is equal and positive if it is greater than addr2 u2.
	ROT 2DUP - >R
	MIN DUP IF
	   >R 
	BEGIN   
	    OVER C@ OVER C@ - IF   SWAP C@ SWAP C@ - R> DROP R> DROP EXIT  THEN     
	    1+ SWAP 1+ SWAP
	    R> 1- DUP >R 0=
	UNTIL R>
	THEN DROP  
	DROP DROP R> NEGATE ;

: M: ( -- )
\G Make a definition similar to : that will be expanded as a macro.
	: LAST @ DUP C@ #32 OR SWAP C! ;

: THROW ( f -- )
\G Will abort if f is nonzero, a preliminary substitute for real THROW.
	ABORT" Throw Exception" ;

M: <= ( n1 n2 -- f )
\G f is true if and only if n1 is less than or equal to n2.
	> 0= ;

M: 0<= ( n1 -- f )
\G f is true if and only if n1 is less than zero.
	0 <= ;

M: BOUNDS ( addr1 n -- addr2 addr1 )
\G Convert address and length to two bounds addresses for DO LOOP 
	OVER + SWAP ;

: WITHIN ( u1 u2  u3 -- f )
\G f is true if u1 is greater or equal to u2 and less than u3
	2 PICK U> ROT ROT U< 0= AND ;

: -TRAILING ( c-addr1 u1 -- c-addr2 u2 )
\G Adjust the length of the string such that trailing spaces are excluded.
	BEGIN
	   2DUP + 1- C@ BL = 
	WHILE
	   1-
	REPEAT ;

M: NIP ( x1 x2 -- x2 )
\G Discard the second item on the stack.
	SWAP DROP ;

\ Added by MHX
: TIMER-RESET	#18 OSCALL ;
: .ELAPSED	#19 OSCALL ;
: CLS		#20 OSCALL ;
: PAGE 		#20 OSCALL ;

\ PART 2: SEARCH ORDER WORDLIST

: GET-ORDER ( -- w1 w2 ... wn n )
\G Return all wordlists in the search order, followed by the count.
	#ORDER @ 0 ?DO  CONTEXT I CELLS + @  LOOP #ORDER @ ;

: SET-ORDER ( w1 w2 ... wn n -- )
\G Set the search order to the n wordlists given on the stack.
	#ORDER ! 0 #ORDER @ 1- DO  CONTEXT I CELLS + !  -1 +LOOP ;

: ALSO ( -- )
\G Duplicate the last wordlist in the search order.
	CONTEXT #ORDER @ CELLS + DUP CELL- @ SWAP ! 1 #ORDER +! ;

: PREVIOUS ( -- )
\G Remove the last wordlist from search order.
	-1 #ORDER +! ;
 
: WORDLIST ( -- wid )
\G Make a new wordlist and give its address.
	HERE 0 , ;

: DEFINITIONS  ( -- )
\G Set the definitions wordlist to the last wordlist in the search order.
	CONTEXT #ORDER @ 1- CELLS + @ CURRENT ! ;

: FORTH ( -- )
\G REplace the last wordlist in the search order with FORTH-WORDLIST
	FORTH-WORDLIST CONTEXT #ORDER @ 1- CELLS + ! ;

WORDLIST 
CONSTANT ROOT-WORDLIST ( -- wid )
\G Minimal wordlist for ONLY

: ONLY ( -- )
\G Set the search order to the minimal wordlist.
	1 #ORDER ! ROOT-WORDLIST CONTEXT ! ;

: VOCABULARY ( -- )
\G Make a definition that will replace the last word in the search order
\G by its wordlist.
	WORDLIST 
	CREATE  ,		\ Make a new wordlist and store it in def.
	DOES>   >R		\ Replace last item in the search order.
		GET-ORDER SWAP DROP R> @ SWAP SET-ORDER ;

\ PART 3: SOME UTILITIES, DUMP .S WORDS
 
: DL ( addr1 -- addr2 )
\G hex/ascii dump in one line of 16 bytes at addr1 addr2 is addr1+16
	BASE @ >R 16 BASE ! CR
	DUP 0 <# # # # # # # # # #> TYPE ." : "
	#16 0 DO  DUP I + C@ 0 <# # # #> TYPE SPACE  LOOP 
	#16 0 DO  DUP I + C@ DUP #127 AND #31 < IF DROP ." ." ELSE EMIT THEN  LOOP 
	#16 + R> BASE ! ;

: DUMP ( addr len --- )
\G Show a hex/ascii dump of the memory block of len bytes at addr  
	#15 + 4 RSHIFT 0 DO  DL  LOOP  DROP ; 

: .S ( --- )
\G Show the contents of the stack.
	DEPTH IF  0 DEPTH 2 - DO I PICK . -1 +LOOP 
     	    ELSE  ." Empty " 
	    THEN ;


: ID. ( nfa --- )
\G Show the name of the word with name field address nfa.
	COUNT #31 AND TYPE SPACE ;

: WORDS ( --- )
\G Show all words in the last wordlist of the search order.
	CONTEXT #ORDER @ 1- CELLS + @ @
	BEGIN
	   DUP 
	WHILE
	   DUP ID.
	   CELL- @
	REPEAT DROP ;


\ SEE instructions 

-7 CONSTANT cellmask 
   VARIABLE ip
   VARIABLE ir

: H. ( u -- ) 
	BASE @ >R HEX   
	0 <#  8 0 DO  #  LOOP [CHAR] $ HOLD  #> TYPE   
	R> BASE ! ;

: SDO_NOOP 	;		: SDO_SWAP	." SWAP " ;	: SDO_ROT	." ROT " ;	: SDO_0=	." 0= " ;
: SDO_NEGATE	." NEGATE " ;	: SDO_UM*	." UM* " ;	: SDO_C@	." C@ " ;	: SDO_@		." @ " ;
: SDO_+		." + " ;	: SDO_AND 	." AND " ;	: SDO_OR	." OR " ;	: SDO_XOR	." XOR " ;
: SDO_U<	." U< " ;	: SDO_<		." < " ;	: SDO_LSHIFT	." LSHIFT " ;	: SDO_RSHIFT	." RSHIFT " ;
: SDO_UM/MOD	." UM/MOD " ;	: SDO_+CY	." +CY " ;	: SDO_SCAN1 	." SCAN1 " ;	: SDO_DROP	." DROP " ;
: SDO_>R	." >R " ;	: SDO_C!A	." C!A " ;	: SDO_!A	." !A " ;	: SDO_DUP	." DUP " ;
: SDO_OVER	." OVER " ;	: SDO_R@	." R@ " ;	: SDO_R>	." R> " ;	: SDO_0		." 0 " ;
: SDO_1		." 1 " ;	: SDO_8		." 8 " ;	

: SDO_LIT	ip @ @  1 CELLS ip +!  . ;
: SDO_SPECIAL 	." SPECIAL " ;

CREATE seers
 ' SDO_NOOP , 	' SDO_SWAP , 	' SDO_ROT , 	' SDO_0= ,
 ' SDO_NEGATE , ' SDO_UM* , 	' SDO_C@ , 	' SDO_@ ,
 ' SDO_+ ,  	' SDO_AND , 	' SDO_OR , 	' SDO_XOR ,
 ' SDO_U< , 	' SDO_< ,  	' SDO_LSHIFT , 	' SDO_RSHIFT ,
 ' SDO_UM/MOD , ' SDO_+CY , 	' SDO_SCAN1 , 	' SDO_SPECIAL ,
 ' SDO_DROP , 	' SDO_>R , 	' SDO_C!A ,	' SDO_!A ,
 ' SDO_DUP , 	' SDO_OVER , 	' SDO_R@ , 	' SDO_R> ,
 ' SDO_0 , 	' SDO_1 , 	' SDO_8 , 	' SDO_LIT ,

\ The virtual SEE machine 
: DO_SEE
	CR ip @ H. 2 SPACES ip @ @ H. 2 SPACES
	ip @ @ ir !  1 CELLS ip +!						\ Instruction fetch.
	ir @ 3 AND 0=  IF ." CALL " ir @ 	      H. EXIT  THEN 		\ Call
	ir @ 3 AND 2 = IF ." JZ   " ir @ cellmask AND H. EXIT  THEN		\ Conditional jump
	ir @ 1 RSHIFT ir !
	#12 0 DO
		 ir @ #31 AND CELLS seers + @ EXECUTE
		 ir @ 5 RSHIFT ir !
	    LOOP
	ir @ 4 AND IF  ." ; "  THEN ;						\ Return

: IDIS ( addr -- )  ip ! BEGIN  DO_SEE  ir @ 4 AND  UNTIL ;
: SEE ( "name" -- ) ' IDIS ;

ROOT-WORDLIST CURRENT !
: FORTH		FORTH ;
: ALSO		ALSO ;
: ONLY		ONLY ;
: PREVIOUS	PREVIOUS ;
: DEFINITIONS	DEFINITIONS ;
: WORDS 	WORDS ;
DEFINITIONS
\ Fill the ROOT wordlist.

: SAVE-SYSTEM ( "ccc" --- )
\G Save the Forth system to a file.
	BL WORD COUNT W/O BIN CREATE-FILE ABORT" Can't open"
	0 HERE 2 PICK WRITE-FILE DROP
	CLOSE-FILE DROP ;

SAVE-SYSTEM forth.img
BYE
