(*
 * LANGUAGE    : ANS Forth with extensions
 * PROJECT     : Forth Environments
 * DESCRIPTION : Longhand division algorithm
 * CATEGORY    : Example 
 * AUTHOR      : Randall Hyde, Art of Assembly
 * LAST CHANGE : March 3, 2008, Marcel Hendrix 
 *)



	NEEDS -miscutil

	REVISION -longhand " Longhand division   Version 0.01 "

	PRIVATES

DOC
(*
   // NumberBits is the number of bits in the Remainder, Quotient, 
   // Divisor, and Dividend variables
	Quotient := Dividend; 
	Remainder := 0;
	for i:= 1 to NumberBits do
	    Remainder:Quotient := Remainder:Quotient SHL 1;
	    if Remainder >= Divisor 
	    then
	       Remainder := Remainder - Divisor;
	       Quotient := Quotient + 1;
	    endif
	endfor
*)
ENDDOC

0. DVALUE DividendHigh	PRIVATE
0. DVALUE DividendLow	PRIVATE -- 128 bits

0. DVALUE QuotientHigh	PRIVATE
0. DVALUE QuotientLow	PRIVATE -- 128 bits

0. DVALUE Remainder	PRIVATE	--  64 bits
0. DVALUE Divisor	PRIVATE	--  64 bits

-- Use DLSHIFT32 with shiftcounts <= 32
: Remainder:Quotient<1 ( -- )
	QuotientHigh D0< 1 AND >R ( cy2)
	QuotientLow  D0< 1 AND >R ( cy1)
	QuotientLow  1 DLSHIFT32 		        TO QuotientLow
	QuotientHigh 1 DLSHIFT32 SWAP R> ( cy1) OR SWAP TO QuotientHigh
	Remainder    1 DLSHIFT32 SWAP R> ( cy2) OR SWAP TO Remainder ; PRIVATE

: UMD/MOD ( uquad ud -- udrem udquot )
	TO Divisor
	TO DividendHigh TO DividendLow
	
	DividendHigh TO QuotientHigh
	DividendLow  TO QuotientLow
	CLEAR Remainder

	#128 0 DO 
		  Remainder:Quotient<1 
		  Remainder Divisor D>= 
		     IF  Divisor -TO Remainder
			 QuotientLow SWAP 1 OR SWAP TO QuotientLow
		  ENDIF
	     LOOP 

	Remainder QuotientLow
	QuotientHigh OR ABORT" result too large" ;

-- UMD/MOD is about 200 times slower than UM/MOD 64/32 unsigned division.
: TESTSPEED ( -- )
	#1000000 LOCAL #iters
	CR ." UMD/MOD : " TIMER-RESET #iters 0 ?DO   0. $22. $44. UMD/MOD 2DROP 2DROP  LOOP .ELAPSED 
	CR ." UM/MOD  : " TIMER-RESET #iters 0 ?DO   0  $22  $44  UM/MOD  2DROP        LOOP .ELAPSED 
	CR #iters U>D (n,3) ." iterations executed." ;

:ABOUT	CR ." Try: ( uquad ud -- udrem udquot ) UMD/MOD " ;

NESTING @ 1 = [IF] 	.ABOUT -longhand CR   [THEN]
			DEPRIVE

                              (* End of Source *)
