(*
 * LANGUAGE    : ANS Forth with extensions
 * PROJECT     : Forth Environments
 * DESCRIPTION : Longhand division algorithm
 * CATEGORY    : Example 
 * AUTHOR      : Terje Mathisen
 * LAST CHANGE : March 3, 2008, Marcel Hendrix 
 *)


	NEEDS -miscutil
	NEEDS -assemble

	REVISION -longhand2 " Longhand division   Version 0.02 "

	PRIVATES

DOC
(*
typedef (single_length unsigned) u
typedef (double_length unsigned) ul

   ul temp = dividend;   // double-length unsigned
   u  result = 0;	// single-length result
   double recip = 1.0 / ((double) divisor)  // Round -> +Inf!

while (temp >= divisor) {
   double dtemp = temp;
   double apprx = dtemp * recip;
   u app = (u) apprx;
   ul backmul = (ul) app * divisor;
// The next line can be skipped if the fp reciprocal is rounded up!
   while (backmul >= temp) { backmul -= divisor; app--; }
   temp -= backmul;
   result += app;
}

At this point 'result' holds the correct answer and 'temp' the remainder 
from the division.

*)
ENDDOC

: UDM*	( lo1 hi1 lo2 hi2 -- uquad )
	LOCALS| hi2 lo2 hi1 lo1 |
	lo1 lo2 UM*  0
	lo2 hi1 UM*  0  UD+c >R
	hi2 lo1 UM*  R> UD+c 
	hi1 hi2 UM*  D+ ;	

CODE UD-b  ( d1 d2 borrow -- d3 borrow )
	eax pop,
	1 b# -> eax rcr,		\ set borrow
	edx pop,		 	\ hi2
	ecx pop,	 		\ lo2 
	edi pop,			\ hi1
	eax pop,			\ lo1
	ecx -> eax sbb, eax push,	\ lo3
	edx -> edi sbb, edi push,	\ hi3
	0 b# -> eax mov, 
	eax  -> eax adc, 
	eax push, 			\ borrow
	ebx jmp,
END-CODE

0. DVALUE backmulHigh	PRIVATE
0. DVALUE backmulLow	PRIVATE -- 128 bits
0. DVALUE RemainderLow	PRIVATE
0. DVALUE RemainderHigh	PRIVATE	-- 128 bits
0. DVALUE tempLow	PRIVATE
0. DVALUE tempHigh	PRIVATE	-- 128 bits

0. DVALUE Quotient	PRIVATE --  64 bits
0. DVALUE Divisor	PRIVATE	--  64 bits
0. DVALUE app		PRIVATE --  64 bits

0e FVALUE recip		PRIVATE

2e #64 FPOW FCONSTANT 2^64 PRIVATE

: corr	  DUP 0< >R  D>F  R> IF  2^64 F+  ENDIF ; PRIVATE ( ud -- ) ( F: -- r )  

: UMD/MOD ( uquad ud -- udrem udquot )
	TO Divisor
	TO RemainderHigh TO RemainderLow
	CLEAR Quotient
	Divisor D>F 1/F TO recip

	BEGIN	
	   RemainderHigh D0<>  RemainderLow Divisor D>= OR
	WHILE
	   RemainderHigh corr 2^64 F* RemainderLow corr F+  recip F* FROUND F>D TO app
	   app divisor UDM* TO backmulHigh TO backmulLow
	   RemainderLow  backmulLow  0  UD-b >R   TO RemainderLow 
	   RemainderHigh backmulHigh R> UD-b DROP TO RemainderHigh
	   app +TO Quotient
	REPEAT

	RemainderLow  Quotient
	RemainderHigh OR ABORT" result too large" ;

-- UMD/MOD is about 5 - 45 times slower than UM/MOD 64/32 unsigned division.
: TESTSPEED ( -- )
	#1000000 LOCAL #iters
	CR ." UMD/MOD : " TIMER-RESET #iters 0 ?DO   0. $22. $44. UMD/MOD 2DROP 2DROP  LOOP .ELAPSED 
	CR ." UMD/MOD : " TIMER-RESET #iters 0 ?DO   $22. 0. $44. UMD/MOD 2DROP 2DROP  LOOP .ELAPSED 
	CR ." UM/MOD  : " TIMER-RESET #iters 0 ?DO   0  $22  $44  UM/MOD  2DROP        LOOP .ELAPSED 
	CR #iters U>D (n,3) ." iterations executed." ;

:ABOUT	CR ." Try: ( uquad ud -- udrem udquot ) UMD/MOD " ;

NESTING @ 1 = [IF] 	.ABOUT -longhand2 CR   [THEN]
			DEPRIVE

                              (* End of Source *)
